@@ -11,11 +11,11 @@ JIRA::Client - An extended interface to JIRA's SOAP API.
 
 =head1 VERSION
 
-Version 0.31
+Version 0.33
 
 =cut
 
-our $VERSION = '0.31';
+our $VERSION = '0.33';
 
 =head1 SYNOPSIS
 
@@ -168,119 +168,144 @@ my %JRA12300_backwards = reverse %JRA12300;
 # These are some helper functions to convert names into ids.
 
 sub _convert_type {
-    my ($self, $hash) = @_;
-    my $type = $hash->{type};
+    my ($self, $type) = @_;
     if ($type =~ /\D/) {
         my $types = $self->get_issue_types();
-        croak "There is no issue type called '$type'.\n"
-            unless exists $types->{$type};
-        $hash->{type} = $types->{$type}{id};
+        return $types->{$type}{id} if exists $types->{$type};
+
+	$types = $self->get_subtask_issue_types();
+        return $types->{$type}{id} if exists $types->{$type};
+
+        croak "There is no issue type called '$type'.\n";
     }
-    return;
+    return $type;
 }
 
 sub _convert_priority {
-    my ($self, $hash) = @_;
-    my $prio = $hash->{priority};
+    my ($self, $prio) = @_;
     if ($prio =~ /\D/) {
         my $prios = $self->get_priorities();
         croak "There is no priority called '$prio'.\n"
             unless exists $prios->{$prio};
-        $hash->{priority} = $prios->{$prio}{id};
+        return $prios->{$prio}{id};
     }
-    return;
+    return $prio;
 }
 
 sub _convert_resolution {
-    my ($self, $hash) = @_;
-    my $resolution = $hash->{resolution};
+    my ($self, $resolution) = @_;
     if ($resolution =~ /\D/) {
         my $resolutions = $self->get_resolutions();
         croak "There is no resolution called '$resolution'.\n"
             unless exists $resolutions->{$resolution};
-        $hash->{resolution} = $resolutions->{$resolution}{id};
+        return $resolutions->{$resolution}{id};
     }
-    return;
+    return $resolution;
 }
 
 sub _convert_security_level {
-    my ($self, $seclevel) = @_;
+    my ($self, $seclevel, $project) = @_;
     if ($seclevel =~ /\D/) {
-        my $seclevels = $self->get_security_levels();
+        my $seclevels = $self->get_security_levels($project);
         croak "There is no security level called '$seclevel'.\n"
             unless exists $seclevels->{$seclevel};
-        $seclevel = $seclevels->{$seclevel}{id};
+        return $seclevels->{$seclevel}{id};
     }
     return $seclevel;
 }
 
+# This routine receives an array with a list of $components specified
+# by RemoteComponent objects, names, and ids. It returns an array of
+# RemoteComponent objects.
+
 sub _convert_components {
-    my ($self, $hash, $key, $project) = @_;
-    my $comps = $hash->{components};
+    my ($self, $components, $project) = @_;
     croak "The 'components' value must be an ARRAY ref.\n"
-       unless ref $comps && ref $comps eq 'ARRAY';
-    my $pcomps;                 # project components
-    foreach my $c (@{$comps}) {
-       next if ref $c;
-       if ($c =~ /\D/) {
-           # It's a component name. Let us convert it into its id.
-	   croak "Cannot convert component names because I don't know for which project.\n"
-	       unless $project;
-           $pcomps = $self->get_components($project) unless defined $pcomps;
-           croak "There is no component called '$c'.\n" unless exists $pcomps->{$c};
-           $c = $pcomps->{$c}{id};
-       }
-       # Now we can convert it into an object.
-       $c = RemoteComponent->new($c);
+       unless ref $components && ref $components eq 'ARRAY';
+    my @converted;
+    my $pcomponents;		# project components
+    foreach my $component (@{$components}) {
+	if (ref $component) {
+	    croak "Invalid object (", ref $component, ") in component's array.\n"
+		unless ref $component eq 'RemoteComponent';
+	    push @converted, $component;
+	} elsif ($component =~ /\D/) {
+	    # It's a component name. Let us convert it into its id.
+	    croak "Cannot convert component names because I don't know for which project.\n"
+		unless $project;
+	    $pcomponents = $self->get_components($project) unless defined $pcomponents;
+	    croak "There is no component called '$component'.\n"
+		unless exists $pcomponents->{$component};
+	    push @converted, RemoteComponent->new($pcomponents->{$component}{id});
+	} else {
+	    push @converted, RemoteComponent->new($component);
+	}
     }
-    return;
+    return \@converted;
 }
 
+# This routine receives an array with a list of $versions specified by
+# RemoteVersion objects, names, and ids. It returns an array of
+# RemoteVersion objects.
+
 sub _convert_versions {
-    my ($self, $hash, $key, $project) = @_;
-    my $versions = $hash->{$key};
-    croak "The '$versions' value must be a ARRAY ref.\n"
+    my ($self, $versions, $project) = @_;
+    croak "The '$versions' value must be an ARRAY ref.\n"
        unless ref $versions && ref $versions eq 'ARRAY';
-    my $pversions;
-    foreach my $v (@{$versions}) {
-       next if ref $v;
-       if ($v =~ /\D/) {
-           # It is a version name. Let us convert it into its id.
-	   croak "Cannot convert version names because I don't know for which project.\n"
-	       unless $project;
-           $pversions = $self->get_versions($project) unless defined $pversions;
-           croak "There is no version called '$v'.\n" unless exists $pversions->{$v};
-           $v = $pversions->{$v}{id};
-       }
-       # Now we can convert it into an object.
-       $v = RemoteVersion->new($v);
+    my @converted;
+    my $pversions;		# project versions
+    foreach my $version (@{$versions}) {
+	if (ref $version) {
+	    croak "Invalid object (", ref $version, ") in versions's array.\n"
+		unless ref $version eq 'RemoteVersion';
+	    push @converted, $version;
+	} elsif ($version =~ /\D/) {
+	    # It is a version name. Let us convert it into its id.
+	    croak "Cannot convert version names because I don't know for which project.\n"
+		unless $project;
+	    $pversions = $self->get_versions($project) unless defined $pversions;
+	    croak "There is no version called '$version'.\n"
+		unless exists $pversions->{$version};
+	    push @converted, RemoteVersion->new($pversions->{$version}{id});
+	} else {
+	    push @converted, RemoteVersion->new($version);
+	}
     }
-    return;
+    return \@converted;
 }
 
+# This routine returns a duedate as a SOAP::Data object with type
+# 'date'. It can generate this from a DateTime object or from a string
+# in the format YYYY-MM-DD.
+
 sub _convert_duedate {
-    my ($self, $hash) = @_;
-    my $duedate = $hash->{duedate};
+    my ($self, $duedate) = @_;
     if (ref $duedate) {
-	return if ref $duedate eq 'SOAP::Data'; # already cast
+	return $duedate if ref $duedate eq 'SOAP::Data'; # already cast
 	croak "duedate fields must be set with DateTime references.\n"
 	    unless ref $duedate eq 'DateTime';
-	$hash->{duedate} = SOAP::Data->type(date => $duedate->strftime('%F'));
-    }
-    elsif (my ($year, $month, $day) = ($duedate =~ /^(\d{4})-(\d{2})-(\d{2})/)) {
+	return SOAP::Data->type(date => $duedate->strftime('%F'));
+    } elsif (my ($year, $month, $day) = ($duedate =~ /^(\d{4})-(\d{2})-(\d{2})/)) {
 	$month >= 1 and $month <= 12
-	    or croak "Invalid duedate ($hash->{duedate}).\n";
-	$hash->{duedate} = SOAP::Data->type(date => $duedate);
+	    or croak "Invalid duedate ($duedate).\n";
+	return SOAP::Data->type(date => $duedate);
+    } else {
+	return $duedate;
     }
-    return;
 }
 
+# This routine receives a hash mapping custom field's ids to
+# values. The ids can be specified by their real id or by their id's
+# numeric suffix (as the 1000 in 'customfield_1000'). Scalar values
+# are substituted by references to arrays containing the original
+# value. The routine returns a hash-ref to another hash with converted
+# keys and values.
+
 sub _convert_custom_fields {
-    my ($self, $hash) = @_;
-    my $custom_fields = $hash->{custom_fields};
+    my ($self, $custom_fields) = @_;
     croak "The 'custom_fields' value must be a HASH ref.\n"
         unless ref $custom_fields && ref $custom_fields eq 'HASH';
-    my %id2values;
+    my %converted;
     while (my ($id, $values) = each %$custom_fields) {
 	my $realid = $id;
         unless ($realid =~ /^customfield_\d+$/) {
@@ -292,9 +317,9 @@ sub _convert_custom_fields {
 
 	# Custom field values must be specified as ARRAYs but we allow for some short-cuts.
 	if (! ref $values) {
-	    $id2values{$realid} = [$values];
+	    $converted{$realid} = [$values];
 	} elsif (ref $values eq 'ARRAY') {
-	    $id2values{$realid} = $values;
+	    $converted{$realid} = $values;
 	} elsif (ref $values eq 'HASH') {
 	    # This is a short-cut for a Cascading select field, which
 	    # must be specified like this: http://tinyurl.com/2bmthoa
@@ -305,9 +330,9 @@ sub _convert_custom_fields {
 		$level_values = [$level_values] unless ref $level_values;
 		if ($level eq '0') {
 		    # The first level doesn't have a colon
-		    $id2values{$realid} = $level_values
+		    $converted{$realid} = $level_values
 		} elsif ($level =~ /^\d+$/) {
-		    $id2values{"$realid:$level"} = $level_values;
+		    $converted{"$realid:$level"} = $level_values;
 		} else {
 		    croak "Invalid cascading field values level spec ($level). It must be a natural number.\n";
 		}
@@ -316,8 +341,7 @@ sub _convert_custom_fields {
 	    croak "Custom field '$id' got a '", ref($values), "' reference as a value.\nValues can only be specified as scalars, ARRAYs, or HASHes though.\n";
 	}
     }
-    $hash->{custom_fields} = \%id2values;
-    return;
+    return \%converted;
 }
 
 my %_converters = (
@@ -336,6 +360,54 @@ foreach my $field (keys %JRA12300) {
     $_converters{$JRA12300{$field}} = $_converters{$field};
 }
 
+# This routine applies all the previous conversions to the $params
+# hash. It returns a reference another hash with converted keys and
+# values, which is the base for invoking the methods createIssue,
+# UpdateIssue, and progressWorkflowAction.
+
+sub _convert_params {
+    my ($self, $params, $project) = @_;
+
+    my %converted;
+
+    # Convert fields' values
+    while (my ($field, $value) = each %$params) {
+	$converted{$field} =
+	    exists $_converters{$field}
+		? $_converters{$field}->($self, $value, $project)
+		    : $value;
+    }
+
+    return \%converted;
+}
+
+# This routine gets a hash produced by _convert_params and flatens in
+# place its Component, Version, and custom_fields fields. It also
+# converts the hash's key according with the %JRA12300 table. It goes
+# a step further before invoking the methods UpdateIssue and
+# progressWorkflowAction.
+
+sub _flaten_components_and_versions {
+    my ($params) = @_;
+
+    # Flaten Component and Version fields
+    for my $field (grep {exists $params->{$_}} qw/components affectsVersions fixVersions/) {
+	$params->{$field} = [map {$_->{id}} @{$params->{$field}}];
+    }
+
+    # Flaten the customFieldValues field
+    if (my $custom_fields = delete $params->{custom_fields}) {
+        while (my ($id, $values) = each %$custom_fields) {
+            $params->{$id} = $values;
+        }
+    }
+
+    # Due to a bug in JIRA we have to substitute the names of some fields.
+    foreach my $field (grep {exists $params->{$_}} keys %JRA12300) {
+	$params->{$JRA12300{$field}} = delete $params->{$field};
+    }
+}
+
 =item B<create_issue> HASH_REF [, SECURITYLEVEL]
 
 Creates a new issue given a hash containing the initial values for its
@@ -400,42 +472,40 @@ You can do it like this:
 
     {customfield_10011 => {'0' => 10031, '1' => 10188}},
 
+Note that the original hash keys and values are completely preserved.
 
 =cut
 
 sub create_issue
 {
-    my ($self, $hash, $seclevel) = @_;
+    my ($self, $params, $seclevel) = @_;
     croak "create_issue requires an argument.\n"
-        unless defined $hash;
+        unless defined $params;
     croak "create_issue's argument must be a HASH ref.\n"
-        unless ref $hash && ref $hash eq 'HASH';
+        unless ref $params && ref $params eq 'HASH';
     for my $field (qw/project summary type/) {
         croak "create_issue's HASH ref must define a '$field'.\n"
-            unless exists $hash->{$field};
+            unless exists $params->{$field};
     }
 
-    # Convert some fields' values
-    foreach my $field (grep {exists $_converters{$_}} keys %$hash) {
-	$_converters{$field}->($self, $hash, $field, $hash->{project});
+    $params = $self->_convert_params($params, $params->{project});
+
+    # Substitute customFieldValues array for custom_fields hash
+    if (my $cfs = delete $params->{custom_fields}) {
+        $params->{customFieldValues} = [map {RemoteCustomFieldValue->new($_, $cfs->{$_})} keys %$cfs];
     }
 
-    # Substitute customFieldValues for custom_fields
-    if (my $cfs = delete $hash->{custom_fields}) {
-        $hash->{customFieldValues} = [map {RemoteCustomFieldValue->new($_, $cfs->{$_})} keys %$cfs];
-    }
-
-    if (my $parent = delete $hash->{parent}) {
+    if (my $parent = delete $params->{parent}) {
 	if (defined $seclevel) {
-	    return $self->createIssueWithParentWithSecurityLevel($hash, _convert_security_level($self, $parent, $seclevel));
+	    return $self->createIssueWithParentWithSecurityLevel($params, $parent, _convert_security_level($self, $seclevel, $params->{project}));
 	} else {
-	    return $self->createIssueWithParent($hash, $parent);
+	    return $self->createIssueWithParent($params, $parent);
 	}
     } else {
 	if (defined $seclevel) {
-	    return $self->createIssueWithSecurityLevel($hash, _convert_security_level($self, $seclevel));
+	    return $self->createIssueWithSecurityLevel($params, _convert_security_level($self, $seclevel, $params->{project}));
 	} else {
-	    return $self->createIssue($hash);
+	    return $self->createIssue($params);
 	}
     }
 }
@@ -443,7 +513,9 @@ sub create_issue
 =item B<update_issue> ISSUE_OR_KEY, HASH_REF
 
 Update a issue given a hash containing the values for its fields. The
-first argument may be an issue key or a RemoteIssue object.
+first argument may be an issue key or a RemoteIssue object. The second
+argument must be a hash-ref specifying the fields's values just like
+documented in the create_issue function above.
 
 This is an easier to use version of the updateIssue API method because
 it accepts the same shortcuts that create_issue does.
@@ -471,36 +543,9 @@ sub update_issue
 
     my ($project) = ($key =~ /^([^-]+)/);
 
-    # Convert some fields' values
-    foreach my $field (grep {exists $_converters{$_}} keys %$params) {
-	$_converters{$field}->($self, $params, $field, $project);
-    }
+    $params = $self->_convert_params($params, $project);
 
-    # Convert RemoteComponent objects into component ids
-    if (my $comps = $params->{components}) {
-        $_ = $_->{id} foreach @$comps;
-    }
-
-    # Convert RemoteVersion objects into version ids
-    for my $field (qw/fixVersions affectsVersions/) {
-        if (my $versions = $params->{$field}) {
-            $_ = $_->{id} foreach @$versions;
-        }
-    }
-
-    # Due to a bug in JIRA we have to substitute the names of some fields.
-    foreach my $field (keys %JRA12300) {
-	if (my $value = delete $params->{$field}) {
-	    $params->{$JRA12300{$field}} = $value;
-	}
-    }
-
-    # Expand the custom_fields hash into the custom fields themselves.
-    if (my $custom_fields = delete $params->{custom_fields}) {
-        while (my ($id, $values) = each %$custom_fields) {
-            $params->{$id} = $values;
-        }
-    }
+    _flaten_components_and_versions($params);
 
     return $self->updateIssue($key, $params);
 }
@@ -518,6 +563,19 @@ sub get_issue_types {
     return $self->{cache}{issue_types};
 }
 
+=item B<get_subtask_issue_types>
+
+Returns a hash mapping the server's sub-task issue type names to the
+RemoteIssueType objects describing them.
+
+=cut
+
+sub get_subtask_issue_types {
+    my ($self) = @_;
+    $self->{cache}{subtask_issue_types} ||= {map {$_->{name} => $_} @{$self->getSubTaskIssueTypes()}};
+    return $self->{cache}{subtask_issue_types};
+}
+
 =item B<get_statuses>
 
 Returns a hash mapping the server's status names to the
@@ -557,17 +615,17 @@ sub get_resolutions {
     return $self->{cache}{resolutions};
 }
 
-=item B<get_security_levels>
+=item B<get_security_levels> PROJECT-KEY
 
-Returns a hash mapping a server's security level names to the
+Returns a hash mapping a project's security level names to the
 RemoteSecurityLevel objects describing them.
 
 =cut
 
 sub get_security_levels {
-    my ($self) = @_;
-    $self->{cache}{seclevels} ||= {map {$_->{name} => $_} @{$self->getSecurityLevels()}};
-    return $self->{cache}{seclevels};
+    my ($self, $project_key) = @_;
+    $self->{cache}{seclevels}{$project_key} ||= {map {$_->{name} => $_} @{$self->getSecurityLevels($project_key)}};
+    return $self->{cache}{seclevels}{$project_key};
 }
 
 =item B<get_custom_fields>
@@ -760,8 +818,8 @@ returned by a previous call to, e.g., C<getIssue>.
 =item C<ACTION> can be either an action I<id> or an action I<name>.
 
 =item C<PARAMS> must be a hash mapping field names to field
-values. This hash accepts the same shortcuts as the argument to
-B<create_issue>.
+values. This hash is treated in the same way as the hash passed to the
+function B<create_issue> above.
 
 =back
 
@@ -780,13 +838,16 @@ And risking to forget to pass some field you can do this:
 =cut
 
 sub progress_workflow_action_safely {
-    my ($self, $key, $action, $params) = @_;
-    my $issue;
-    if (ref $key) {
-        $issue = $key;
+    my ($self, $issue, $action, $params) = @_;
+    my $key;
+    if (ref $issue) {
+	croak "progress_workflow_action_safely's first argument must be a RemoteIssue reference.\n"
+	    unless ref $issue eq 'RemoteIssue';
         $key   = $issue->{key};
+    } else {
+	$key   = $issue;
+	$issue = undef;
     }
-    my ($project) = (split /-/, $key)[0];
     $params = {} unless defined $params;
     ref $params and ref $params eq 'HASH'
         or croak "progress_workflow_action_safely's third arg must be a HASH-ref\n";
@@ -816,48 +877,15 @@ sub progress_workflow_action_safely {
         if (exists $issue->{$id}) {
             $params->{$id} = $issue->{$id} if defined $issue->{$id};
         }
-        else {
-            foreach my $cf (@{$issue->{customFieldValues}}) {
-                if ($cf->{customfieldId} eq $id) {
-                    $params->{$id} = $cf->{values};
-                    last;
-                }
-            }
-            # NOTE: It's not a problem if we can't find a missing
-            # parameter in the issue. It will simply stay undefined.
-        }
+	# NOTE: It's not a problem if we can't find a missing
+	# parameter in the issue. It will simply stay undefined.
     }
 
-    # Convert some fields' values
-    foreach my $field (grep {exists $_converters{$_}} keys %$params) {
-	$_converters{$field}->($self, $params, $field, $project);
-    }
+    my ($project) = ($key =~ /^([^-]+)/);
 
-    # Convert RemoteComponent objects into component ids
-    if (my $comps = $params->{components}) {
-        $_ = $_->{id} foreach @$comps;
-    }
+    $params = $self->_convert_params($params, $project);
 
-    # Convert RemoteVersion objects into version ids
-    for my $field (qw/fixVersions affectsVersions/) {
-        if (my $versions = $params->{$field}) {
-            $_ = $_->{id} foreach @$versions;
-        }
-    }
-
-    # Due to a bug in JIRA we have to substitute the names of some fields.
-    foreach my $field (keys %JRA12300) {
-	if (my $value = delete $params->{$field}) {
-	    $params->{$JRA12300{$field}} = $value;
-	}
-    }
-
-    # Expand the custom_fields hash into the custom fields themselves.
-    if (my $custom_fields = delete $params->{custom_fields}) {
-        while (my ($id, $values) = each %$custom_fields) {
-            $params->{$id} = $values;
-        }
-    }
+    _flaten_components_and_versions($params);
 
     return $self->progressWorkflowAction($key, $action, $params);
 }
@@ -1397,7 +1425,7 @@ L<http://search.cpan.org/dist/JIRA-Client>
 
 =head1 COPYRIGHT & LICENSE
 
-Copyright 2009-2011 CPqD, all rights reserved.
+Copyright 2009-2012 CPqD, all rights reserved.
 
 This program is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
